/*
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <Camera.h>
#include <Adafruit_GFX.h>
#include <Adafruit_ILI9341.h>
#include <Servo.h>

static Servo sv; /**< Servo object */

#include <SDHCI.h>
#include <stdio.h>  /* for sprintf */

#define TFT_DC 9
#define TFT_CS 10

// 画面内位置
#define XP_HDR 170
#define YP_HDR 0
#define XP_FIL 170
#define YP_FIL 30
#define XP_ERR 0
#define YP_ERR 290
#define XP_ISO 0
#define YP_ISO 130
#define XP_SHUT 0
#define YP_SHUT 160
#define XP_WB 0
#define YP_WB 190
#define XP_FILE 0
#define YP_FILE 220
#define XP_CAP 0
#define YP_CAP 250
#define XP_JPG 170
#define YP_JPG 60

#define QTINIT 80 //JPEG品質初期値

Adafruit_ILI9341 display = Adafruit_ILI9341(TFT_CS, TFT_DC);

SDClass  theSD;

int button4 = HIGH;
int button5 = HIGH;
int button6 = HIGH;
int button7 = HIGH;

int take_picture_count = 0;
bool hdrmode = false;  //HDRモード

//            IRCF  NoIR  700nm 740nm 810nm
int fpos[] = {2160, 1820, 1460, 1110, 780};
char* filter[] = {"IRCF_", "NoIR_", "700nm", "740nm", "810nm"};

char* wbname[] = {"Automatic   ", "Incandescent", "Fluorescent ", "Daylight    ", "Flash       ", "Cloudy      ", "Shade       "};
CAM_WHITE_BALANCE    wbmode[] = {CAM_WHITE_BALANCE_AUTO,
                                 CAM_WHITE_BALANCE_INCANDESCENT,
                                 CAM_WHITE_BALANCE_FLUORESCENT,
                                 CAM_WHITE_BALANCE_DAYLIGHT,
                                 CAM_WHITE_BALANCE_FLASH,
                                 CAM_WHITE_BALANCE_CLOUDY,
                                 CAM_WHITE_BALANCE_SHADE
                                };
int wbpos = 0 ; // ホワイトバランスモード
int fpt = 0 ; //フィルタ位置
int dtime = 500 ;
int qt = QTINIT ;
float imgsize[5];
int iso[5] ;
int exposure[5] ;

void printError(enum CamErr err)
{
  display.setTextColor(ILI9341_RED, ILI9341_BLACK);  display.setTextSize(1);
  display.setCursor(XP_ERR, YP_ERR );
  display.print("Error: ");
  switch (err)
  {
    case CAM_ERR_NO_DEVICE: display.println("No Device"); break;
    case CAM_ERR_ILLEGAL_DEVERR: display.println("Illegal device error"); break;
    case CAM_ERR_ALREADY_INITIALIZED: display.println("Already initialized"); break;
    case CAM_ERR_NOT_INITIALIZED: display.println("Not initialized"); break;
    case CAM_ERR_NOT_STILL_INITIALIZED: display.println("Still picture not initialized"); break;
    case CAM_ERR_CANT_CREATE_THREAD: display.println("Failed to create thread"); break;
    case CAM_ERR_INVALID_PARAM: display.println("Invalid parameter"); break;
    case CAM_ERR_NO_MEMORY: display.println("No memory"); break;
    case CAM_ERR_USR_INUSED: display.println("Buffer already in use"); break;
    case CAM_ERR_NOT_PERMITTED: display.println("Operation not permitted");  break;
    default: break;
  }
}

void CamCB(CamImage img) {
  if (img.isAvailable()) {
    img.convertPixFormat(CAM_IMAGE_PIX_FMT_RGB565);
    display.drawRGBBitmap(0, 0 /* 開始座標 */
                          , (uint16_t*)img.getImgBuff() /* 画像データ */
                          , 160 , 120 ); /* 横幅、縦幅  */
  }
}

void setup() {
  CamErr err;
  pinMode(PIN_D02, INPUT_PULLUP);
  pinMode(PIN_D03, INPUT_PULLUP);
  pinMode(PIN_D04, INPUT_PULLUP);
  pinMode(PIN_D07, INPUT_PULLUP);

  sv.attach(PIN_D05); // サーボ
  sv.writeMicroseconds(fpos[fpt]); //IRCF位置に移動

  display.begin(40000000); //　液晶ディスプレイの開始
  display.setRotation(0); // ディスプレイの向きを設定

  err = theCamera.begin(1, CAM_VIDEO_FPS_30, 160, 120, CAM_IMAGE_PIX_FMT_YUV422, 7 ); // カメラの開始

  if (err != CAM_ERR_SUCCESS)
  {
    printError(err);
  }
  err = theCamera.setAutoWhiteBalanceMode(CAM_WHITE_BALANCE_AUTO); //ホワイトバランスをautoに設定
  if (err != CAM_ERR_SUCCESS)
  {
    printError(err);
  }
  err = theCamera.setAutoISOSensitivity(true);//ISO感度をautoに設定
  if (err != CAM_ERR_SUCCESS)
  {
    printError(err);
  }
  err = theCamera.setAutoExposure(true);//露光時間をautoに設定
  if (err != CAM_ERR_SUCCESS)
  {
    printError(err);
  }
  err = theCamera.setHDR(CAM_HDR_MODE_OFF); //HDRモードをOFFに設定
  if (err != CAM_ERR_SUCCESS)
  {
    printError(err);
  }
    err = theCamera.setStillPictureImageFormat(1280, 960, CAM_IMAGE_PIX_FMT_JPG, 4); //JPEG静止画設定
  if (err != CAM_ERR_SUCCESS)
  {
    printError(err);
  }

  theCamera.setJPEGQuality(qt);
  theCamera.startStreaming(true, CamCB); // カメラのストリーミングを開始

  //LCD初期描画
  display.fillScreen(ILI9341_BLACK);
  display.setTextColor(ILI9341_YELLOW, ILI9341_BLACK);  display.setTextSize(2);
  display.setCursor(XP_HDR, YP_HDR );
  display.println("SDR");
  display.setTextColor(ILI9341_CYAN, ILI9341_BLACK);  display.setTextSize(2);
  display.setCursor(XP_FIL, YP_FIL );
  display.print(filter[fpt]);
  display.setTextColor(ILI9341_RED, ILI9341_BLACK);  display.setTextSize(2);
  display.setCursor(XP_JPG, YP_JPG );
  display.printf("Q:%d", qt);
  display.setTextColor(ILI9341_MAGENTA, ILI9341_BLACK);  display.setTextSize(2);
  display.setCursor(XP_WB, YP_WB );
  display.print(wbname[wbpos]);

  //SDカードマウント確認
  while (!theSD.begin())
  {
    display.setTextColor(ILI9341_RED, ILI9341_BLACK);  display.setTextSize(2);
    display.setCursor(XP_ERR, YP_ERR );
    display.print("Insert SD card.");
  }
  display.setCursor(XP_ERR, YP_ERR );
  display.print("               ");

  filecheck(); //ファイルネーム番号初期値チェック「
}

void change_filter(int fpt)
{
  sv.writeMicroseconds(fpos[fpt]);
  display.setTextColor(ILI9341_CYAN, ILI9341_BLACK);  display.setTextSize(2);
  display.setCursor(XP_FIL, YP_FIL );
  display.print(filter[fpt]);
  if (fpt > 1)
  {
    theCamera.setColorEffect(CAM_COLOR_FX_BW );
  } else {
    theCamera.setColorEffect(CAM_COLOR_FX_NONE );
  }
  delay(1000);
}

void filecheck()
{
  char filename[17] = {0};
  char filename1[17] = {0};
  char filename2[17] = {0};
  char filename3[17] = {0};
  char filename4[17] = {0};

  sprintf(filename, "PICT%03d_%s.JPG", take_picture_count, filter[0]);
  sprintf(filename1, "PICT%03d_%s.JPG", take_picture_count, filter[1]);
  sprintf(filename2, "PICT%03d_%s.JPG", take_picture_count, filter[2]);
  sprintf(filename3, "PICT%03d_%s.JPG", take_picture_count, filter[3]);
  sprintf(filename4, "PICT%03d_%s.JPG", take_picture_count, filter[4]);

  while (
    theSD.exists(filename ) ||
    theSD.exists(filename1 ) ||
    theSD.exists(filename2 ) ||
    theSD.exists(filename3 ) ||
    theSD.exists(filename4 )   )
  {
    take_picture_count++;
    sprintf(filename, "PICT%03d_%s.JPG", take_picture_count, filter[0]);
    sprintf(filename1, "PICT%03d_%s.JPG", take_picture_count, filter[1]);
    sprintf(filename2, "PICT%03d_%s.JPG", take_picture_count, filter[2]);
    sprintf(filename3, "PICT%03d_%s.JPG", take_picture_count, filter[3]);
    sprintf(filename4, "PICT%03d_%s.JPG", take_picture_count, filter[4]);
  }
  display.setTextColor(ILI9341_BLUE, ILI9341_BLACK);  display.setTextSize(2);
  display.setCursor(XP_FILE, YP_FILE );
  display.print("Start No:");
  display.print(take_picture_count);
}

void capture()
{
  display.setTextColor(ILI9341_RED, ILI9341_BLACK);  display.setTextSize(2);
  display.setCursor(XP_ERR, YP_ERR );
  display.print("                    ");
  char filename[17] = {0};
  sprintf(filename, "PICT%03d_%s.JPG", take_picture_count, filter[0]);
  for ( auto fpt = 0 ; fpt < 5 ; fpt++)
  {
    change_filter(fpt);
    display.setTextColor(ILI9341_RED, ILI9341_BLACK);  display.setTextSize(2);
    display.setCursor(XP_JPG, YP_JPG );
    display.printf("Q:%d", qt);
    CamImage img = theCamera.takePicture();
    getinfo(fpt);
    if (!img.isAvailable())
    {
      qt = qt - 5 ;
      theCamera.setJPEGQuality(qt);
      display.setTextColor(ILI9341_RED, ILI9341_BLACK);  display.setTextSize(2);
      display.setCursor(XP_JPG, YP_JPG );
      display.printf("Q:%d", qt);
      CamImage img = theCamera.takePicture();
    }
    qt = QTINIT ;
    if (img.isAvailable())
    {
      display.setTextColor(ILI9341_RED, ILI9341_BLACK);  display.setTextSize(2);
      display.setCursor(XP_ERR, YP_ERR );
      imgsize[fpt] = img.getImgSize();
      display.printf("RAM:%.0f/%.0f[KB]", imgsize[fpt]  / 1024.0, img.getImgBuffSize() / 1024.0);
      sprintf(filename, "PICT%03d_%s.JPG", take_picture_count, filter[fpt]);
      display.setTextColor(ILI9341_BLUE, ILI9341_BLACK);  display.setTextSize(2);
      display.setCursor(XP_FILE, YP_FILE );
      display.print(filename);
      File myFile = theSD.open(filename, FILE_WRITE);
      myFile.write(img.getImgBuff(), img.getImgSize());
      myFile.close();
    }
    else
    {
      display.setTextColor(ILI9341_RED, ILI9341_BLACK);  display.setTextSize(2);
      display.setCursor(XP_ERR, YP_ERR );
      display.print("Save Failed:");
      display.print(filter[fpt]);
    }
  }

  sprintf(filename, "PICT%03d_info.txt", take_picture_count);
  File myFile = theSD.open(filename, FILE_WRITE);
  myFile.print("HDR mode:");
  myFile.println(theCamera.getHDR());
  myFile.print("white balance:");
  myFile.println(wbname[wbpos]);
  myFile.println("" );
  for ( auto i = 0 ; i < 5 ; i++)
  {
    myFile.print("ISO ");
    myFile.print(filter[i]);
    myFile.print(":" );
    myFile.println(iso[i]);
    myFile.print("SHUT ");
    myFile.print(filter[i]);
    myFile.print(":" );
    myFile.println(exposure[i]);
    myFile.print("Size " );
    myFile.print(filter[i]);
    myFile.print(":" );
    myFile.println(imgsize[i]);
    myFile.println("");
  }
  myFile.close();

  change_filter(0);
  take_picture_count++;
}

void getinfo(int pos)
{
  iso[pos] = theCamera.getISOSensitivity();// ISO値取得
  exposure[pos] = theCamera.getAbsoluteExposure(); //露光時間取得

  display.setTextColor(ILI9341_WHITE, ILI9341_BLACK);  display.setTextSize(2);
  display.setTextWrap(false) ;
  display.setCursor(XP_ISO, YP_ISO );
  display.print("ISO:");
  display.setCursor(XP_ISO + 60, YP_ISO );
  display.print(iso[pos] );
  display.print("     ");
  display.setCursor(XP_SHUT, YP_SHUT );
  display.print("SHUT:");
  display.setCursor(XP_SHUT + 60, YP_SHUT );
  display.print(exposure[pos]);
  display.print("     ");
  display.setTextSize(2);
  display.setCursor(XP_HDR + 40, YP_HDR );
  display.print(theCamera.getHDR());
}

void loop()
{
  CamErr err;
  int button4_input = digitalRead(PIN_D04);
  int button5_input = digitalRead(PIN_D02);
  int button6_input = digitalRead(PIN_D03);
  int button7_input = digitalRead(PIN_D07);

  // Still Capture
  if (button4_input != button4) {
    if (button4_input == LOW) {
      // push  on
      display.setTextColor(ILI9341_RED, ILI9341_BLACK);  display.setTextSize(2);
      display.setCursor(XP_CAP, YP_CAP);
      display.print("Capture");
      capture();
      display.setCursor(XP_CAP, YP_CAP);
      display.print("       ");
    } else {
      /// push off
    }
    button4 = button4_input;
  }

  //White Balance mode
  if (button5_input != button5) {
    if (button5_input == LOW) {
      // push  on
      wbpos++;
      if (wbpos > 6 )
      {
        wbpos = 0;
      }
      display.setTextColor(ILI9341_MAGENTA, ILI9341_BLACK);  display.setTextSize(2);
      display.setCursor(XP_WB, YP_WB );
      display.print(wbname[wbpos]);
      theCamera.setAutoWhiteBalanceMode(wbmode[wbpos]);
      delay(200);
    } else {
      /// push off
    }
    button5 = button5_input;
  }

  //filter change
  if (button6_input != button6) {
    if (button6_input == LOW) {
      // push  on
      fpt++;
      if (fpt > 4 )
      {
        fpt = 0;
      }
      change_filter(fpt);
    } else {
      /// push off
    }
    button6 = button6_input;
  }

  // HDR mode set
  if (button7_input != button7) {
    if (button7_input == LOW) {
      // push  on
      hdrmode = !hdrmode ;
      if (hdrmode == true) {
        err = theCamera.setHDR(CAM_HDR_MODE_ON);
        if (err != CAM_ERR_SUCCESS)
        {
          printError(err);
        }
        display.setTextColor(ILI9341_YELLOW, ILI9341_BLACK);  display.setTextSize(2);
        display.setCursor(XP_HDR, YP_HDR);
        display.print("HDR");
        delay(dtime);
      }
      else if ( hdrmode == false) {
        err = theCamera.setHDR(CAM_HDR_MODE_OFF);
        if (err != CAM_ERR_SUCCESS)
        {
          printError(err);
        }
        display.setTextColor(ILI9341_YELLOW, ILI9341_BLACK);  display.setTextSize(2);
        display.setCursor(XP_HDR, YP_HDR );
        display.print("SDR");
        delay(dtime);
      }
    } else {
      /// push off
    }
    button7 = button7_input;
  }
  getinfo(0);

}
